import AST.*;
import java.util.*;
import Lexer.*;
import java.io.*;


public class Compiler {
    
      // compile must receive an input with an character less than 
      // p_input.lenght
   public Program compile( char []input, PrintWriter outError ) {
        
       symbolTable = new SymbolTable();
       error = new CompilerError( lexer, new PrintWriter(outError) );
       lexer = new Lexer(input, error);
       error.setLexer(lexer);
        
        
       Program p = null;
       try {
          lexer.nextToken();
          if ( lexer.token == Symbol.EOF )
             error.show("Unexpected EOF");
          p = program();
          if ( lexer.token != Symbol.EOF ) {
             p = null;
             error.show("EOF expected");
          }
       } catch ( Exception e ) {
              // the below statement prints the stack of called methods.
              // of course, it should be removed if the compiler were 
              // a production compiler.
            //e.printStackTrace();
          p = null;
       }
       
       //# corrija: a chamada de createResult deve estar na
       // class Comp, NO aqui. A chamada foi colocada aqui
       // porque este mtodo (compile)
       // retorna null mesmo se houve sucesso na compilao.
       /*if ( error.wasAnErrorSignalled() ) {
          p = null;
          Comp.createResult("0");
       }
       else 
          Comp.createResult("1");
       */
       return p;   
   }
    
   
   private Program program() {
        
      classDec();
      while ( lexer.token == Symbol.CLASS ) 
         classDec();
      return new Program();
   }

   private void classDec() {
      if ( lexer.token != Symbol.CLASS ) 
         error.show("'class' expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.IDENT )
         error.show(CompilerError.identifier_expected);
      String className = lexer.getStringValue();
      lexer.nextToken();
      if ( lexer.token == Symbol.SUBCLASSOF ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
         String superclassName = lexer.getStringValue();
         lexer.nextToken();
      }
      if ( lexer.token == Symbol.PROC )
         error.show("Attempt to declare a method outside the private or public section");
      if ( lexer.token == Symbol.VAR ) 
         error.show("Attempt to declare instance variables outside the private or public sections");
      if ( lexer.token == Symbol.PRIVATE ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.COLON )
            error.show(": expected");
         lexer.nextToken();
         InstanceVariableList instVarList = new InstanceVariableList();
         instVarDecMethodDec(instVarList);
         while ( lexer.token == Symbol.VAR ||
                 lexer.token == Symbol.PROC ) 
            instVarDecMethodDec(instVarList);
      }
      
      if ( lexer.token == Symbol.PUBLIC ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.COLON )
            error.show(": expected");
         lexer.nextToken();
         methodDec();
         while ( lexer.token == Symbol.PROC ) 
            methodDec();
      }
      
      if ( lexer.token != Symbol.END )
         error.show("'end' expected");
      lexer.nextToken();
   }
   
   
   private void instVarDecMethodDec(InstanceVariableList instVarList) {
      if ( lexer.token == Symbol.VAR ) 
         instVarDecList(instVarList);
      else if ( lexer.token == Symbol.PROC )
         methodDec();
      else
         error.show("'var' or 'proc' expected");
   }
   
   
   private void instVarDecList( InstanceVariableList instVarList ) {
        // lexer.token == MySym.VAR
      lexer.nextToken();
      instVarDec(instVarList);
      while ( lexer.token == Symbol.IDENT )
         instVarDec(instVarList);
   }
           
        
   private void instVarDec( InstanceVariableList instVarList ) {
      Vector localList = new Vector();
      InstanceVariable p;
      while ( true ) {
         if ( lexer.token != Symbol.IDENT ) 
            error.show(CompilerError.identifier_expected);
         else {
            p = new InstanceVariable(lexer.getStringValue());
            instVarList.addElement(p);
            localList.addElement(p);
            lexer.nextToken();
         }
         if ( lexer.token == Symbol.COMMA )
            lexer.nextToken();
         else
            break;
      }
      if ( lexer.token != Symbol.COLON )
         error.show(": expected");
      else
         lexer.nextToken();
      Type t = type();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      else
         lexer.nextToken();

      Enumeration e = localList.elements();
      while ( e.hasMoreElements() ) {
         p = (InstanceVariable ) e.nextElement();
         p.setType(t);
      }
   }


   private void methodDec() {
      if ( lexer.token != Symbol.PROC )
         if ( lexer.token == Symbol.VAR ) 
            error.show("Attempt to declare instance variables in the public section");
         else
            error.show("'proc' expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.IDENT )
         error.show(CompilerError.identifier_expected);
      String methodName = lexer.getStringValue();
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR )
         error.show("( expected");
      lexer.nextToken();
      if ( lexer.token == Symbol.IDENT )
         formalParamDec();
      if ( lexer.token != Symbol.RIGHTPAR )
         error.show(") or ; expected");
      lexer.nextToken();
      if ( lexer.token == Symbol.COLON ) {
         lexer.nextToken();
         type();
      }
      if ( lexer.token == Symbol.VAR ) 
         localDec();
      compositeStatement();
   }

        
   private LocalVarList localDec() {
        // LocalVarDec ::= "var" VarDecList 
        // VarDecList ::= VarDecList2  { VarDecList2  }
        
      LocalVarList localVarList = new LocalVarList();
        // eat token "var"
      lexer.nextToken();
        
      varDec(localVarList);
      while ( lexer.token == Symbol.IDENT )
         varDec(localVarList);
      return localVarList;
   }


   private void varDec( LocalVarList localVarList ) {
       //  VarDecList2 ::= Ident { ',' Ident } ':' Type ';'

      Vector tempVarList = new Vector();
        
      while ( true ) {
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
            // name of the identifier
         String name = lexer.getStringValue();
         lexer.nextToken();
         //# corrija
         tempVarList.addElement( new Variable(name) );
          
         if ( lexer.token == Symbol.COMMA ) 
            lexer.nextToken();
         else 
            break;
      }

      if ( lexer.token != Symbol.COLON ) 
         error.show(": expected");
      else
         lexer.nextToken();
          // get the type
      Type typeVar = type();
        
      Variable v;
      Enumeration e = tempVarList.elements();
      while ( e.hasMoreElements() ) {
           // add type to the variable
         v = (Variable ) e.nextElement();
         v.setType(typeVar);
           // add variable to the list of variable declarations
         localVarList.addElement( v );
      }
        
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      else
         lexer.nextToken();  
   }

   
          
   private ParamList formalParamDec() {

      ParamList paramList = new ParamList();
      paramDec(paramList);
      while ( lexer.token == Symbol.SEMICOLON ) {
         lexer.nextToken();
         paramDec(paramList);
      }
      return paramList;
   }

   private void paramDec( ParamList paramList ) {

      Vector lastVarList = new Vector();
        
      while ( true ) {
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
            // name of the identifier
         String name = lexer.getStringValue();
         lexer.nextToken();
            // variable does not have a type yet
         Parameter v = new Parameter(name);
            // list of the last variables declared. They don't have types yet
         lastVarList.addElement(v);
          
         if ( lexer.token == Symbol.COMMA ) 
            lexer.nextToken();
         else 
            break;
      }

      if ( lexer.token != Symbol.COLON )
         error.show(": expected");
      else
         lexer.nextToken();
          // get the type
      Type typeVar = type();
        
      Parameter v;
      Enumeration e = lastVarList.elements();
      while ( e.hasMoreElements() ) {
           // add type to the variable
         v = (Parameter ) e.nextElement();
         v.setType(typeVar);
            // add v to the list of parameter declarations
         paramList.addElement(v);
      }
        
   }
          
        

    private Type type() {
        Type result;
        
        switch ( lexer.token ) {
            case Symbol.INTEGER :
               result = Type.integerType;
               break;
            case Symbol.BOOLEAN :
               result = Type.booleanType;
               break;
            case Symbol.STRING :
               result = Type.stringType;
               break;
            case Symbol.IDENT :
               //# corrija: faa uma busca na TS para buscar a classe
               // IDENT deve ser uma classe.
               result = null;
               break;
            default :
              error.show("Type expected");
              result = Type.undefinedType;
        }
        lexer.nextToken();
        return result;
    }
    
        
        

    private void compositeStatement() {
        
        if ( lexer.token != Symbol.BEGIN )  
            error.show("'begin' expected");
        else
          lexer.nextToken();
        statementList();
        if ( lexer.token != Symbol.END )
          error.show("'end' expected");
        else
          lexer.nextToken();
    }



   private void statementList() {
       int tk;
          // statements always begin with an identifier, if, read, write, ...
       while ( (tk = lexer.token) != Symbol.END && 
               tk != Symbol.ELSE &&
               tk != Symbol.ENDIF ) 
          statement();
   }
            
   private void statement() {
           
      switch (lexer.token) {
         case Symbol.IDENT :
         case Symbol.SUPER :
         case Symbol.SELF :
            assignmentMessageSendStatement();
            break;
         case Symbol.IF :
            ifStatement();
            break;
         case Symbol.WHILE :
            whileStatement();
            break;
         case Symbol.RETURN :
            returnStatement();
            break;
         case Symbol.READ :
            readStatement();
            break;
         case Symbol.WRITE :
            writeStatement();
            break;
         case Symbol.LOOP :
            loopStatement();
            break;
         case Symbol.BREAK :
            breakStatement();
            break;
         case Symbol.SEMICOLON :
            nullStatement();
            break;
         default :
            error.show("Statement expected");
      }
   }
    
   private Statement assignmentMessageSendStatement() {
      String methodName, variableName;
      ExprList exprList;
      Statement result = null;
      /*
         there are six possibilities:
            super.id()
            self.id()
            self.id = expr
            self.id.id()
            id = expr
            id.id()
      */
      
      switch ( lexer.token ) {
         case Symbol.SUPER :
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");            
            lexer.nextToken();
            if ( lexer.token != Symbol.IDENT )
               error.show(CompilerError.identifier_expected);
            methodName = lexer.getStringValue();
            lexer.nextToken();
            exprList = getRealParameters();
            //# corrija
            // result = new MessageSendStatement( 
            //     new MessageSendToSuper( pointer to class, pointer to method, exprList) );
            break;
         case Symbol.SELF :
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");
            lexer.nextToken();
            if ( lexer.token != Symbol.IDENT )
               error.show(CompilerError.identifier_expected);
            String ident = lexer.getStringValue();
            lexer.nextToken();
            switch ( lexer.token ) {
               case Symbol.ASSIGN :
                  lexer.nextToken();
                  Expr anExpr = expr();
                  //# corrija
                  /* result = new MessageSendStatement( new AssignmentStatement( pointer to instance variable,
                        anExpr) ); */
                  break;
               case Symbol.DOT :
                  lexer.nextToken();
                  if ( lexer.token != Symbol.IDENT )
                     error.show(CompilerError.identifier_expected);
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                        new MessageSendToVariable( pointer to variable,
                           pointer to method, exprList) );  */
                  break;
               case Symbol.LEFTPAR :
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                       new MessageSendToSelf( pointer to method, exprList ) ); */
                  break;
               default :
                  error.show(CompilerError.identifier_expected);
            }
            break;
         case Symbol.IDENT :
            variableName = lexer.getStringValue();
            lexer.nextToken();
            switch ( lexer.token ) {
               case Symbol.ASSIGN : 
                  lexer.nextToken();
                  Expr anExpr = expr();
                  //# corrija
                  /* result = new AssignmentStatement( pointer to variable,
                        anExpr ); */
                  break;
               case Symbol.DOT :
                  lexer.nextToken();
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                         new MessageSendToVariable( pointer to variable,
                               pointer to method ) ); */
                  break;
               default :
                  error.show(". or = expected");
            }
            break;
         default :
            error.show("'self', 'super', or identifier expected");
      }
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
      return result;
   }

    
   private ExprList getRealParameters() {
      ExprList anExprList = null;
      
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      if ( startExpr(lexer.token) )
         anExprList = exprList();
      if ( lexer.token != Symbol.RIGHTPAR )
         error.show(") expected");
      lexer.nextToken();
      return anExprList;
   }
    
   private void ifStatement() {
        
      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.THEN ) 
         error.show("then expected");
      lexer.nextToken();
      statementList();
      if ( lexer.token == Symbol.ELSE ) {
         lexer.nextToken();
         statementList();
      }
      if ( lexer.token != Symbol.ENDIF )
         error.show("endif expected");
      lexer.nextToken();
   }
            
   private void whileStatement() {

      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.DO )
         error.show("do expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.BEGIN ) 
         statement();
      else {
         lexer.nextToken();
         statementList();
         if ( lexer.token != Symbol.END )
            error.show("'end' expected");
         lexer.nextToken();
      }
   }
    
    
   private void returnStatement() {
        
      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
          
            
   private void readStatement() {
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      while ( true ) {
         if ( lexer.token == Symbol.SELF ) {
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");
            lexer.nextToken();
         }
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
         
         String name = (String ) lexer.getStringValue();
         lexer.nextToken();
         if ( lexer.token == Symbol.COMMA )
            lexer.nextToken();
         else
            break;
      }
      
      if ( lexer.token != Symbol.RIGHTPAR ) 
         error.show(") expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
    
    
   private void writeStatement() {
        
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      exprList();
      if ( lexer.token != Symbol.RIGHTPAR ) 
          error.show(") expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
        
   private void loopStatement() {
      lexer.nextToken();
      statementList();
      if ( lexer.token != Symbol.END )
         error.show("'end' expected");
      lexer.nextToken();
   }

   private void breakStatement() {
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
   
   private void nullStatement() {
      lexer.nextToken();
   }
    
   private ExprList exprList() {
    
      ExprList anExprList = new ExprList();
      anExprList.addElement(expr());
      while ( lexer.token == Symbol.COMMA ) {
         lexer.nextToken();
         anExprList.addElement(expr());
      }
      return anExprList;
   }



   private Expr expr() {
    
      Expr left = simpleExpr();
      int op = lexer.token;
      if ( op == Symbol.EQ || op == Symbol.NEQ || 
           op == Symbol.LE || op == Symbol.LT ||
           op == Symbol.GE || op == Symbol.GT  ) {
         lexer.nextToken();
         Expr right = simpleExpr();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }


   private Expr simpleExpr() {
      int op;
      
      Expr left = term();
      while ( (op = lexer.token) == Symbol.PLUS || op == Symbol.MINUS || 
              op == Symbol.OR ) {
         lexer.nextToken();
         Expr right = term();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }


   private Expr term() {
      int op;
      
      Expr left = signalFactor();
      while ( (op = lexer.token) == Symbol.MULT || op == Symbol.DIV || 
              op == Symbol.AND ) {
         lexer.nextToken();
         Expr right = signalFactor();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }
   
   private Expr signalFactor() {
      int op;
      if ( (op = lexer.token) == Symbol.PLUS || op == Symbol.MINUS ) {
         lexer.nextToken();
         return new SignalExpr(op, factor());
      }
      else
         return factor();
   }
   


    private Expr factor() {
      /*
        Factor ::= Number | "true" | "false" | "self" [ "." Id ]
           | Id | MessageSend | '(' Expr ')' | "not" Factor | "nil"
      */
      
      Expr e;
      
      switch ( lexer.token ) {
         case Symbol.NUMBER :
            return number();
         case Symbol.TRUE :
           lexer.nextToken();
           return BooleanExpr.True;
         case Symbol.FALSE :
           lexer.nextToken();
           return BooleanExpr.False;
         case Symbol.LITERALSTRING :
           String literalString = lexer.getLiteralStringValue();
           lexer.nextToken();
           return new LiteralStringExpr(literalString);
         case Symbol.LEFTPAR :
            lexer.nextToken();
            e = expr();
            if ( lexer.token != Symbol.RIGHTPAR ) 
               error.show(") expected");
            lexer.nextToken();
            return new ParenthesisExpr(e);
         case Symbol.NOT :
            lexer.nextToken();
            e = expr();
            return new UnaryExpr( e, Symbol.NOT );
         case Symbol.NIL :
            lexer.nextToken();
            return new NilExpr();
         default :
            String variableName, methodName;   
            ExprList exprList;
               /* there are seven cases to consider :
                     x
                     self.x
                     self
                     super.m()
                     x.m()
                     A.new()
                     self.m()
                     self.x.m();
                     
                  in which x is an instance variable and m is a method
               */
            switch ( lexer.token ) {
               case Symbol.IDENT :
                  variableName = lexer.getStringValue();
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT ) {
                     //# corrija
                     //return new VariableExpr(pointer to variable);
                  }
                  else {
                     lexer.nextToken();
                     switch ( lexer.token ) {
                        case Symbol.IDENT :
                           methodName = lexer.getStringValue();
                           lexer.nextToken();
                           exprList = getRealParameters();
                           //#  corrija
                           /* return new MessageSendToVariable(
                                 pointer to variable, pointer to method) */
                           break;
                        case Symbol.NEW :
                           lexer.nextToken();
                           if ( lexer.token != Symbol.LEFTPAR ) 
                              error.show("( expected");
                           lexer.nextToken();
                           if ( lexer.token != Symbol.RIGHTPAR )
                              error.show(") expected");
                           lexer.nextToken();
                           //# corrija
                           // return new MessageSendNew(pointer to class) 
                           break;
                        default :
                           error.show(CompilerError.identifier_expected);
                     }
                  }
                  break;
               case Symbol.SUPER :
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT )
                     error.show(". expected");
                  lexer.nextToken();
                  if ( lexer.token != Symbol.IDENT )
                     error.show(CompilerError.identifier_expected);
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //#  corrija
                  /* return new MessageSendToSuper(
                         pointer to method) */
                  break;
               case Symbol.SELF :
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT ) {
                     //# corrija
                     // return new SelfExpr();
                  }
                  else {
                     lexer.nextToken();
                     if ( lexer.token != Symbol.IDENT )
                        error.show(CompilerError.identifier_expected);
                        // it may be method name or an instance variable
                     String ident = lexer.getStringValue();
                     lexer.nextToken();
                     switch ( lexer.token ) {
                        case Symbol.LEFTPAR : 
                           // expression of the kind self.m()
                           //# corrija
                           exprList = getRealParameters();
                           /*
                           return new MessageSendToSelf( pointer to method,
                              exprList );
                           */
                           break;
                        case Symbol.DOT :
                           // expression of the kind self.x.m()
                           //# corrija
                           lexer.nextToken();
                           if ( lexer.token != Symbol.IDENT )
                              error.show(CompilerError.identifier_expected);
                           methodName = lexer.getStringValue();
                           lexer.nextToken();
                           exprList = getRealParameters();
                           /*
                           return new MessageSendToVariable( 
                              pointer to instance variable ident,
                              pointer to method,
                              exprList );
                           */
                        default :
                           // expression of the kind self.x
                           //# corrija
                           /*
                           return new VariableExpr( pointer to instance 
                              variable ident )
                           */
                     }
                        
                  }
                  break;
               default : 
                  error.show(CompilerError.identifier_expected);
               }
               return null;
         }
            
   }

    
   private NumberExpr number() {
        
      NumberExpr e = null;
        
          // the number value is stored in lexer.getToken().value as an object of Integer.
          // Method intValue returns that value as an value of type int.
      int value = lexer.getNumberValue();
      lexer.nextToken();
      return new NumberExpr( value );
   }
    
    
   private boolean startExpr( int aToken ) {
    
      return lexer.token == Symbol.IDENT || 
             lexer.token == Symbol.NUMBER ||
             lexer.token == Symbol.FALSE ||
             lexer.token == Symbol.TRUE ||
             lexer.token == Symbol.SELF || 
             lexer.token == Symbol.SUPER ||
             lexer.token == Symbol.LEFTPAR ||
             lexer.token == Symbol.NOT ||
             lexer.token == Symbol.NIL ||
             lexer.token == Symbol.LITERALSTRING;
           
   }
    
    
   private SymbolTable symbolTable;
   private Lexer lexer;
   private CompilerError error;
  
}
