import AST.*;
import java.io.*;
import java.util.*;

public class RComp {
    
    public static void main( String []args ) {
        
        File file;
        FileReader stream;
        int numChRead;
        Program program;
        
        String originalClass, toBeCreatedClass;
        Hashtable methodTable, instanceVariableTable;
        Vector methodArray, instanceVariableArray;
        String inputFileName, outputFileName;
        CompilerOptions compilerOptions;
        boolean doingRefactoring;
        
        String resultFileName; 
        
        if ( args.length < 1 )  {
           printUsage();
        }
        else {
           originalClass = toBeCreatedClass = "";
           methodTable = new Hashtable();
           instanceVariableTable = new Hashtable();
           methodArray = new Vector();
           instanceVariableArray = new Vector();
           compilerOptions = new CompilerOptions();
           
           boolean soleOption = false;
           boolean hasOptions = args[0].charAt(0) == '-';
           int i = 0;
           while ( i < args.length && args[i].charAt(0) == '-' ) {
              String s = args[i];
              if ( s.length() != MaxSizeOption || i + 1 >= args.length )
                 printUsage();
              if ( soleOption ) {
                 error("Wrong combination of options in command line");
              }
              switch( s.charAt(1) ) {
                 case 'o' :
                    if ( originalClass.length() != 0 )
                       printUsage();
                    i++;
                    originalClass = args[i];
                    compilerOptions.setExtractClass(true);
                    break;
                 case 'n' :
                    if ( toBeCreatedClass.length() != 0 )
                       printUsage();
                    i++;
                    toBeCreatedClass = args[i];
                    compilerOptions.setExtractClass(true);
                    break;
                 case 'm' :
                    i++;
                    if ( methodTable.put(args[i], "") != null )
                       error("Duplicate method name");
                    methodArray.addElement(args[i]);
                    compilerOptions.setExtractClass(true);
                    break;
                 case 'i' :
                    i++;
                    if ( instanceVariableTable.put(args[i], "") != null )
                       error("Duplicate instance variable name");
                    instanceVariableArray.addElement(args[i]);
                    compilerOptions.setExtractClass(true);
                    break;
                 case 'c' :
                    compilerOptions.setCount(true);
                       // this should be the only option in command line
                    soleOption = true;
                    break;
                 case 'p' :
                    compilerOptions.setOutputInterface(true);
                       // this should be the only option in command line
                    soleOption = true;
                    break;
                 default :
                    printUsage();
              }
              i++;
           }
           if ( i >= args.length )
              printUsage();
           inputFileName = args[i++];
           if ( i < args.length )
              outputFileName = args[i];
           else {
              outputFileName = inputFileName;
              int lastIndex;
              if ( (lastIndex = outputFileName.lastIndexOf('.')) == -1 )
                lastIndex = outputFileName.length();
              StringBuffer sb = new StringBuffer(outputFileName.substring(0, lastIndex));
              if ( compilerOptions.getOutputInterface() )
                 sb.append(".i");
              else
                 sb.append(".c");
              outputFileName = sb.toString();
           }
           
           if ( hasOptions && compilerOptions.getExtractClass() ) {
              if ( originalClass.length() == 0 ||
                   toBeCreatedClass.length()== 0 ||
                   methodTable.size() == 0 )
                 error("Refactoring data is wrong. Missing class names or methods to be moved.");
           }
           
           /*
           System.out.println("input: " + inputFileName);
           System.out.println("output: " + outputFileName);
           if ( hasOptions ) {
              System.out.print("original class = " + originalClass + "\n" +
                 "new class = " + toBeCreatedClass + "\n" + 
                 "methods : ");
              Enumeration e = methodTable.keys();
              while ( e.hasMoreElements() )
                 System.out.print( (String ) e.nextElement() + "  " );
              System.out.println();
              System.out.print("instance variables : ");
              e = instanceVariableTable.keys();
              while ( e.hasMoreElements() )
                 System.out.print( (String ) e.nextElement() + "  " );
              System.out.println();
           }
           */
           
           file = new File(inputFileName);
           int index = inputFileName.indexOf(".s");
           if ( index <= 0 && 
                (index = inputFileName.indexOf(".S")) <= 0) {
              System.out.println("Error: file does not .s extension");
              System.exit(1);
           }
           resultFileName = inputFileName.substring(0, index) + ".txt";

           if ( ! file.exists() || ! file.canRead() ) {
             System.out.println("Either the file " + inputFileName + " does not exist or it cannot be read");
             createResult(resultFileName, "0");
             return ;
           }
           try { 
             stream = new FileReader(file);  
            } catch ( FileNotFoundException e ) {
                System.out.println("Something wrong: file does not exist anymore");
                createResult(resultFileName, "0");
                throw new RuntimeException();
            }
                // one more character for '\0' at the end that will be added by the
                // compiler
            char []input = new char[ (int ) file.length() + 1 ];
            
            try {
              numChRead = stream.read( input, 0, (int ) file.length() );
            } catch ( IOException e ) {
                System.out.println("Error reading file " + inputFileName);
                createResult(resultFileName, "0");
                return ;
            }
                
            if ( numChRead != file.length() ) {
                System.out.println("Read error");
                createResult(resultFileName, "0");
                return ;
            }
            try {
              stream.close();
            } catch ( IOException e ) {
                createResult(resultFileName, "0");
                System.out.println("Error in handling the file " + inputFileName);
                return ;
            }
                

            Compiler compiler = new Compiler();
            
                
            FileOutputStream  outputStream;
            try { 
               outputStream = new FileOutputStream(outputFileName);
            } catch ( IOException e ) {
                System.out.println("File " + outputFileName + " was not found");
                createResult(resultFileName, "0");
                return ;
            }
            PrintWriter printWriter = new PrintWriter(outputStream);
            program = null;
              // the generated code goes to a file and so are the errors
            program  = compiler.compile(input, new PrintWriter(System.out) );
            
            if ( program == null ) {
               createResult(resultFileName, "0");
            }
            else {
               PW pw = new PW();
               pw.set(printWriter);
               //# corrija
               /*
               if ( compilerOptions.getOutputInterface() ) 
                  program.genInterfaces(pw);
               else if ( compilerOptions.getExtractClass() ) {
                  Refactoring refactoring = new Refactoring(program);
                  refactoring.extractClass( originalClass, toBeCreatedClass,
                       methodArray, instanceVariableArray );
                     // genSimples just prints the AST in Simples
                  program.genSimples(pw);
               }
               else
                  // the genC methods will test if the statistics should
                  // be generated.
                  program.genC(pw, compilerOptions);
               */   
               createResult(resultFileName, "1");
               if ( printWriter.checkError() ) {
                  System.out.println("There was an error in the output");
               }
            }
            printWriter.close();
        }
    }
    
   public static boolean createResult(String resultFileName, String contents) {
    
      //System.out.println("result = " + resultFileName);
      PrintWriter out;
	  FileOutputStream outputStream = null;
      try {
         outputStream = new FileOutputStream(resultFileName);
      } catch ( FileNotFoundException  e) {
         System.out.println("Could not create " + resultFileName);
         return false;
      }
      out = new PrintWriter(outputStream);
      
      out.println(contents);
      out.close();
      return true;
   }
   
   public static void printUsage() {
      System.out.println("Usage:\n   Comp [options] input [output]");
      System.out.println("input is the file to be compiled");
      System.out.println("output is the file where the generated code will be stored");
      System.out.println("If options o, n, m, and i are included, the refactoring Extract Class should be applied (item 2 refactorings)");
      System.out.println("options include\n" + 
         "   -o name, the name of the old class\n" +
         "   -n name, the name of the new class to be created\n" + 
         "   -m name, name of a method to be moved to the new class\n" +
         "   -i name, name of an instance variable to be moved to the new class\n" + 
         "   -p       print the interfaces of all classes in the output (item 4 homework)\n" +
         "   -c       produce output in C with code to count the number of times each method is called (item 2 homework)\n");
      /*
        Example: suppose we want to create a class Student from a class Person.
        Methods getCourse, setCourse, getNumber, and setNumber should be 
        moved from Person to Student. Instance variables course and
        number should be moved too. This can be achieved by calling
        this program with the following command line:
           java Comp -o Person -n Student -m getCourse -m setCourse
                     -m getNumber -m setNumber -i course -i number ok-chk07.s
                      new-ok-chk07.s
        The classes are in a file "ok-chk07.s". A new file 
        "new-ok-chk07.s" will be created with all the classes of
        "ok-chk07.s". But in this new file class Person was 
        split into a new class Person and a new class Student.
       */
       System.exit(1);
   }
   
   public static void error( String message ) {
      System.out.println(message);
      System.exit(1);
   }
   
   static private final int MaxSizeOption = 2;
}
        